/* (C) 2011,2016 Yana */

#include "device.h"
#include "yanamips.h"

typedef signed char y_int8_t;
typedef signed short y_int16_t;
typedef signed int y_int32_t;
typedef unsigned char y_uint1_t;
typedef unsigned char y_uint5_t;
typedef unsigned char y_uint7_t;
typedef unsigned char y_uint8_t;
typedef unsigned short y_uint16_t;
typedef unsigned int y_uint26_t;
typedef unsigned int y_uint32_t;

union {
	y_uint32_t w;
	y_uint8_t b[4];
} m_main[4096] = {
#include "m_main.h"
};

y_uint8_t cin;

y_uint32_t r_gpr[32];

y_uint32_t r_badvaddr; /* 8 */
y_uint32_t r_status; /* 12 */
y_uint32_t r_cause; /* 13 */
y_uint32_t r_epc; /* 14 */

y_uint32_t r_pc;
y_uint1_t r_bd;
y_uint32_t r_next;

y_uint32_t handle(y_uint7_t exccode) {
	r_cause = r_cause & ~YANAMIPS_CAUSE_EXCCODE_MASK | exccode;
	if ((r_status & YANAMIPS_STATUS_EXL) == 0) {
		r_status |= YANAMIPS_STATUS_EXL;
		if (r_bd) {
			r_cause |= YANAMIPS_CAUSE_BD;
			r_epc = r_pc - 4;
		} else {
			r_cause &= ~YANAMIPS_CAUSE_BD;
			r_epc = r_pc;
		}
	}
	return YANAMIPS_VECTOR;
}

y_uint8_t hlscore(y_uint8_t reset, y_uint8_t c) {
	y_uint8_t cout = 0;
	y_uint32_t code;
	y_uint26_t c2i;
	y_uint5_t c4s;
	y_uint5_t c4t;
	y_uint16_t c4i;
	y_uint5_t c6s;
	y_uint5_t c6t;
	y_uint5_t c6d;
	y_uint5_t c6h;
	y_uint1_t bd;
	y_uint32_t next;
	y_uint32_t i;
	y_uint32_t j;
	if (reset != 0) {
		r_gpr[0] = 0;
		r_cause = 0;
		r_status = 0;
		r_pc = 0;
		r_bd = 0;
		r_next = r_pc + 4;
		return 0;
	}
	if (c != 0) {
		cin = c;
		r_cause |= DEVICE_INTERRUPT_CONS << YANAMIPS_CAUSE_IP_SHIFT;
	}
	if ((r_cause >> YANAMIPS_CAUSE_IP_SHIFT & (r_status & YANAMIPS_STATUS_IM_MASK) >> YANAMIPS_STATUS_IM_SHIFT) != 0 && (r_status & (YANAMIPS_STATUS_EXL | YANAMIPS_STATUS_IE)) == YANAMIPS_STATUS_IE) {
		r_pc = handle(YANAMIPS_CAUSE_EXCCODE_INT);
		r_bd = 0;
		r_next = r_pc + 4;
	}
	bd = 0;
	next = r_next + 4;
	/* ToDo: check r_pc */
	code = m_main[r_pc >> 2].w;
	c2i = code & 0x03ffffff;
	c4s = code >> 21 & 0x1f;
	c4t = code >> 16 & 0x1f;
	c4i = code & 0xffff;
	c6s = code >> 21 & 0x1f;
	c6t = code >> 16 & 0x1f;
	c6d = code >> 11 & 0x1f;
	c6h = code >> 6 & 0x1f;
	switch (code >> 26) {
	case 0x00:
		switch (code & 0x3f) {
		case 0x00: /* sll */
			/* ToDo: check c6h */
			i = r_gpr[c6t] << c6h;
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x02: /* srl */
			/* ToDo: check c6h */
			i = r_gpr[c6t] >> c6h;
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x03: /* sra */
			/* ToDo: check c6h */
			i = (y_int32_t)r_gpr[c6t] >> c6h;
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x24: /* and */
			i = r_gpr[c6s] & r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x25: /* or */
			i = r_gpr[c6s] | r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x26: /* xor */
			i = r_gpr[c6s] ^ r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x08: /* jr */
			/* ToDo: check r_bd */
			bd = 1;
			next = r_gpr[c6s];
			break;
		case 0x09: /* jalr */
			/* ToDo: check r_bd */
			i = r_next + 4;
			j = r_gpr[c6s];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			bd = 1;
			next = j;
			break;
		case 0x0c: /* syscall */
			r_next = handle(YANAMIPS_CAUSE_EXCCODE_SYS);
			next = r_next + 4;
			break;
		case 0x21: /* addu */
			i = r_gpr[c6s] + r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x23: /* subu */
			i = r_gpr[c6s] - r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x2a: /* slt */
			i = (y_int32_t)r_gpr[c6s] < (y_int32_t)r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x2b: /* sltu */
			i = r_gpr[c6s] < r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		default:
			/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
			break;
		}
		break;
	case 0x02: /* j */
		/* ToDo: check r_bd */
		bd = 1;
		next = r_next & 0xf0000000 | c2i << 2;
		break;
	case 0x03: /* jal */
		/* ToDo: check r_bd */
		r_gpr[31] = r_next + 4;
		bd = 1;
		next = r_next & 0xf0000000 | c2i << 2;
		break;
	case 0x04: /* beq */
		/* ToDo: check r_bd */
		bd = 1;
		if (r_gpr[c4s] == r_gpr[c4t]) {
			next = r_next + ((y_int16_t)c4i << 2);
		}
		break;
	case 0x05: /* bne */
		/* ToDo: check r_bd */
		bd = 1;
		if (r_gpr[c4s] != r_gpr[c4t]) {
			next = r_next + ((y_int16_t)c4i << 2);
		}
		break;
	case 0x09: /* addiu */
		i = r_gpr[c4s] + (y_int16_t)c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0a: /* slti */
		i = (y_int32_t)r_gpr[c4s] < (y_int16_t)c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0b: /* sltiu */
		i = r_gpr[c4s] < (y_int16_t)c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0c: /* andi */
		i = r_gpr[c4s] & c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0d: /* ori */
		i = r_gpr[c4s] | c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0e: /* xori */
		i = r_gpr[c4s] ^ c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0f: /* lui */
		i = c4i << 16;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x10:
		switch (c6s) {
		case 0x00: /* mfc0 */
			switch (c6d) {
			case 8:
				i = r_badvaddr;
				break;
			case 12:
				i = r_status;
				break;
			case 13:
				i = r_cause;
				break;
			case 14:
				i = r_epc;
				break;
			default:
				/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
				break;
			}
			if (c6t != 0) {
				r_gpr[c6t] = i;
			}
			break;
		case 0x04: /* mtc0 */
			switch (c6d) {
			case 8:
				break;
			case 12:
				r_status = r_status & ~(YANAMIPS_STATUS_IM_MASK | YANAMIPS_STATUS_EXL | YANAMIPS_STATUS_IE) | r_gpr[c6t] & (YANAMIPS_STATUS_IM_MASK | YANAMIPS_STATUS_EXL | YANAMIPS_STATUS_IE);
				break;
			case 13:
				r_cause = r_cause & ~(YANAMIPS_CAUSE_IP1 | YANAMIPS_CAUSE_IP0) | r_gpr[c6t] & (YANAMIPS_CAUSE_IP1 | YANAMIPS_CAUSE_IP0);
				break;
			case 14:
				r_epc = r_gpr[c6t];
				break;
			default:
				/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
				break;
			}
			break;
		case 0x10: /* eret */
			/* ToDo: check r_bd */
			r_status &= ~YANAMIPS_STATUS_EXL;
			r_next = r_epc;
			next = r_next + 4;
			break;
		default:
			/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
			break;
		}
		break;
	case 0x20: /* lb */
		switch (r_gpr[c4s] + (y_int16_t)c4i) {
		case DEVICE_CONS_PUTGETCHAR:
			i = cin;
			if (c4t != 0) {
				r_gpr[c4t] = i;
			}
			cin = 0;
			r_cause &= ~(DEVICE_INTERRUPT_CONS << YANAMIPS_CAUSE_IP_SHIFT);
			break;
		case DEVICE_CONS_HALT:
			/* ToDo: halt */
			break;
		default:
#if 0
			i = m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[r_gpr[c4s] + (y_int16_t)c4i & 3];
#else
			i = m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[~(r_gpr[c4s] + (y_int16_t)c4i) & 3];
#endif
			if (c4t != 0) {
				r_gpr[c4t] = i;
			}
			break;
		}
		break;
	case 0x23: /* lw */
		if ((r_gpr[c4s] + (y_int16_t)c4i & 3) != 0) {
			r_badvaddr = r_gpr[c4s] + (y_int16_t)c4i;
			r_next = handle(YANAMIPS_CAUSE_EXCCODE_ADEL);
			next = r_next + 4;
		} else {
			i = m_main[(r_gpr[c4s] + (y_int16_t)c4i) >> 2].w;
			if (c4t != 0) {
				r_gpr[c4t] = i;
			}
		}
		break;
	case 0x28: /* sb */
		switch (r_gpr[c4s] + (y_int16_t)c4i) {
		case DEVICE_CONS_PUTGETCHAR:
			cout = r_gpr[c4t];
			break;
		case DEVICE_CONS_HALT:
			/* ToDo: halt */
			break;
		default:
#if 0
			m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[r_gpr[c4s] + (y_int16_t)c4i & 3] = r_gpr[c4t];
#else
			m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[~(r_gpr[c4s] + (y_int16_t)c4i) & 3] = r_gpr[c4t];
#endif
			break;
		}
		break;
	case 0x2b: /* sw */
		if ((r_gpr[c4s] + (y_int16_t)c4i & 3) != 0) {
			r_badvaddr = r_gpr[c4s] + (y_int16_t)c4i;
			r_next = handle(YANAMIPS_CAUSE_EXCCODE_ADES);
			next = r_next + 4;
		} else {
			m_main[(r_gpr[c4s] + (y_int16_t)c4i) >> 2].w = r_gpr[c4t];
		}
		break;
	default:
		 /* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
		break;
	}
	r_pc = r_next;
	r_bd = bd;
	r_next = next;
	return cout;
}
